"""Provides all the data related to the hardware."""
from itertools import product

RESOLUTIONS = [
    "1152x768",
    "1280x854",
    "1440x960",
    "2880x1920",
    "1024x768",
    "1152x864",
    "1280x960",
    "1400x1050",
    "1600x1200",
    "2048x1536",
    "3200x2400",
    "1280x768",
    "1280x1024",
    "2560x2048",
    "1280x720",
    "1365x768",
    "1600x900",
    "1920x1080",
    "1280x800",
    "1440x900",
    "1680x1050",
    "1920x1200",
    "2560x1600",
    "3840x2400",
]

SCREEN_SIZES = [
    "14″",
    "12.1″",
    "12″",
    "14.4″",
    "15″",
    "15.7″",
    "13.3″",
    "13″",
    "17″",
    "15.4″",
    "14.1″",
    "16″",
    "27″",
    "29″",
    "34″",
    "32″",
    "40″",
]

CPU = [
    "AMD Ryzen 7 1800X",
    "AMD Ryzen 7 1700",
    "AMD Ryzen™ Threadripper™",
    "Intel® Core i3",
    "Intel® Core i5",
    "Intel® Core i7",
    "Intel® Core i9",
    "Apple M1",
    "Apple M1 Pro",
    "Apple M1 Max",
    "Apple M2",
]

RAM_TYPES = [
    "SDRAM",
    "DDR",
    "DDR2",
    "DDR3",
    "DDR4",
    "DDR5",
]

RAM_SIZES = [
    "4GB",
    "8GB",
    "12GB",
    "16GB",
    "32GB",
    "64GB",
    "128GB",
]

GENERATION = [
    "2nd Generation",
    "3rd Generation",
    "4th Generation",
    "5th Generation",
    "6th Generation",
    "7th Generation",
    "8th Generation",
    "9th Generation",
]

CPU_CODENAMES = [
    "Ivytown",
    "Haswell",
    "Fortville",
    "Devil's Canyon",  # noqa: Q003
    "Valley Island",
    "Broadwell",
    "Bay Trail",
    "Skylake",
    "Orchid Island",
    "Bear Ridge",
    "Cannonlake",
]

HDD_SSD_MANUFACTURERS = [
    "Western Digital",
    "Seagate",
    "Samsung",
    "Intel",
    "Micron",
    "Kingston",
    "SanDisk",
]

_CAPACITY = [
    "64GB SSD",
    "128GB SSD",
    "256GB SDD",
    "512GB SSD",
    "1TB SSD",
    "2TB SSD",
    "4TB SSD",
    "256GB HDD",
    "512GB HDD",
    "1TB HDD",
    "2TB HDD",
    "4TB HDD",
    "8TB HDD",
]

HDD_SSD = [f"{i[0]} {i[1]}" for i in product(HDD_SSD_MANUFACTURERS, _CAPACITY)]

GRAPHICS = [
    "AMD Radeon PRO WX 8200",
    "AMD Radeon Pro W5700",
    "AMD Radeon RX 5500 XT",
    "AMD Radeon RX 560",
    "AMD Radeon RX 5600 XT",
    "AMD Radeon RX 570",
    "AMD Radeon RX 5700",
    "AMD Radeon RX 5700 XT",
    "AMD Radeon RX 580",
    "AMD Radeon RX 590",
    "AMD Radeon RX 6500 XT",
    "AMD Radeon RX 6600 XT",
    "AMD Radeon RX 6700 XT",
    "AMD Radeon RX 6750 XT",
    "AMD Radeon RX 6800",
    "AMD Radeon RX 6800 XT",
    "AMD Radeon RX 6900 XT",
    "AMD Radeon RX 6950 XT",
    "AMD Radeon RX Vega 56",
    "AMD Radeon RX Vega 64",
    "AMD Radeon VII",
    "Intel® HD Graphics 3000",
    "Intel® HD Graphics 4000",
    "Intel® HD Graphics 4400",
    "Intel® HD Graphics 4600",
    "Intel® HD Graphics 5000",
    "Intel® HD Graphics 520",
    "Intel® HD Graphics 5300 ",
    "Intel® HD Graphics 5500",
    "Intel® HD Graphics 6000",
    "Intel® HD Graphics 615",
    "Intel® HD Graphics 620",
    "Intel® Iris™ Graphics 5100",
    "Intel® Iris™ Graphics 550",
    "Intel® Iris™ Graphics 6100",
    "Intel® Iris™ Pro Graphics 5200",
    "Intel® Iris™ Pro Graphics 580",
    "Intel® Iris™ Pro Graphics 6200",
    "Nvidia GTX 1050",
    "Nvidia GTX 1050 Ti",
    "Nvidia GTX 1060",
    "Nvidia GTX 1070",
    "Nvidia GTX 1070 Ti",
    "Nvidia GTX 1080",
    "Nvidia GTX 1080 Ti",
    "Nvidia GTX 1650",
    "Nvidia GTX 1650 SUPER",
    "Nvidia GTX 1660",
    "Nvidia GTX 1660 SUPER",
    "Nvidia GTX 1660 Ti",
    "Nvidia GTX 960",
    "Nvidia GTX 980",
    "Nvidia GTX 980 Ti",
    "Nvidia Quadro RTX A4000",
    "Nvidia Quadro RTX A5000",
    "Nvidia Quadro RTX A6000",
    "Nvidia RTX 2060",
    "Nvidia RTX 2060 SUPER",
    "Nvidia RTX 2070",
    "Nvidia RTX 2070 SUPER",
    "Nvidia RTX 2080",
    "Nvidia RTX 2080 SUPER",
    "Nvidia RTX 2080 Ti",
    "Nvidia RTX 3050",
    "Nvidia RTX 3050",
    "Nvidia RTX 3050 Ti",
    "Nvidia RTX 3060",
    "Nvidia RTX 3060",
    "Nvidia RTX 3060 Ti",
    "Nvidia RTX 3070",
    "Nvidia RTX 3070",
    "Nvidia RTX 3070 Ti",
    "Nvidia RTX 3080",
    "Nvidia RTX 3080",
    "Nvidia RTX 3080 Ti",
    "Nvidia RTX 3090",
    "Nvidia RTX 3090 Ti",
    "Nvidia RTX 4090",
    "Nvidia RTX Titan",
    "Nvidia Titan Pascal",
    "Nvidia Titan V",
]

MANUFACTURERS = [
    "Apple",
    "Acer",
    "Dell",
    "ASUS",
    "VAIO",
    "Lenovo",
    "HP",
    "Toshiba",
    "Sony",
    "Samsung",
    "Fujitsu",
    "Xiomi",
]

PHONE_MODELS = [
    "iPhone SE",
    "iPhone X",
    "iPhone XS",
    "iPhone 11",
    "iPhone 11 Pro",
    "iPhone 11 Pro Max",
    "iPhone 12",
    "iPhone 12 Pro",
    "iPhone 12 Pro Max",
    "iPhone 13",
    "iPhone 13 Pro",
    "iPhone 13 Pro Max",
    "iPhone 14",
    "iPhone 14 Plus",
    "iPhone 14 Pro",
    "iPhone 14 Pro Max",
    "iPhone 15 Pro",
    "iPhone 15 Pro Max",
    "Nothing Phone",
    "Samsung Galaxy S22 Ultra",
    "Samsung Galaxy S22 Plus",
    "Samsung Galaxy Fold 4",
    "Samsung Galaxy Z Flip 4",
    "Xiaomi Redmi Note 11",
    "Xiaomi 12 Pro",
    "Google Pixel 6",
    "Google Pixel 6 Pro",
    "Google Pixel 7",
    "Google Pixel 7 Pro",
    "Vivo X80 Pro",
    "OnePlus 10 Pro",
]



"""Address module.

This module contains provider Address() and other utils which represent
data related to location, such as street name, city etc.
"""

import typing as t

from mimesis.datasets import (
    CALLING_CODES,
    CONTINENT_CODES,
    COUNTRY_CODES,
    SHORTENED_ADDRESS_FMT,
)
from mimesis.enums import CountryCode
from mimesis.providers.base import BaseDataProvider

__all__ = ["Address"]


class Address(BaseDataProvider):
    """Class for generate fake address data.

    This object provides all the data related to
    geographical location.
    """

    class Meta:
        name = "address"
        datafile = f"{name}.json"

    @staticmethod
    def _dd_to_dms(num: float, _type: str) -> str:
        """Converts decimal number to DMS format.

        :param num: Decimal number.
        :param _type: Type of number.
        :return: Number in DMS format.
        """
        direction = ""
        if _type == "lg":
            direction = "W" if num < 0 else "E"
        elif _type == "lt":
            direction = "S" if num < 0 else "N"

        num = abs(num)
        degrees = int(num)
        part = num - degrees
        minutes = int(part * 60)
        seconds = 3600 * part - 60 * minutes
        seconds = round(seconds, 3)

        return f"{degrees}º{minutes}'{seconds:.3f}\"{direction}"

    def street_number(self, maximum: int = 1400) -> str:
        """Generates a random street number.

        :param maximum: Maximum value.
        :return: Street number.
        """
        return str(self.random.randint(1, maximum))

    def street_name(self) -> str:
        """Generates a random street name.

        :return: Street name.
        """
        street_names: list[str] = self._extract(["street", "name"])
        return self.random.choice(street_names)

    def street_suffix(self) -> str:
        """Generates a random street suffix.

        :return: Street suffix.
        """
        suffixes: list[str] = self._extract(["street", "suffix"])
        return self.random.choice(suffixes)

    def address(self) -> str:
        """Generates a random full address.

        :return: Full address.
        """
        fmt: str = self._extract(["address_fmt"])

        st_num = self.street_number()
        st_name = self.street_name()

        if self.locale in SHORTENED_ADDRESS_FMT:
            return fmt.format(
                st_num=st_num,
                st_name=st_name,
            )

        if self.locale == "ja":
            return fmt.format(
                self.random.choice(self._extract(["city"])),
                # Generate a list of random integers
                # in n of 3, from 1 to 100.
                *self.random.randints(n=3, a=1, b=100),
            )

        return fmt.format(
            st_num=st_num,
            st_name=st_name,
            st_sfx=self.street_suffix(),
        )

    def state(self, abbr: bool = False) -> str:
        """Generates a random administrative district of the country.

        :param abbr: Return ISO 3166-2 code.
        :return: Administrative district.
        """
        key = "abbr" if abbr else "name"
        states: list[str] = self._extract(["state", key])
        return self.random.choice(states)

    def region(self, *args: t.Any, **kwargs: t.Any) -> str:
        """Generates a random region.

        An alias for :meth:`~.state()`.
        """
        return self.state(*args, **kwargs)

    def province(self, *args: t.Any, **kwargs: t.Any) -> str:
        """Generates a random province.

        An alias for :meth:`~.state()`.
        """
        return self.state(*args, **kwargs)

    def federal_subject(self, *args: t.Any, **kwargs: t.Any) -> str:
        """Generates a random federal_subject (Russia).

        An alias for :meth:`~.state()`.
        """
        return self.state(*args, **kwargs)

    def prefecture(self, *args: t.Any, **kwargs: t.Any) -> str:
        """Generates a random prefecture.

        An alias for :meth:`~.state()`.
        """
        return self.state(*args, **kwargs)

    def postal_code(self) -> str:
        """Generates a postal code for current locale.

        :return: Postal code.
        """
        return self.random.generate_string_by_mask(self._extract(["postal_code_fmt"]))

    def zip_code(self) -> str:
        """Generates a zip code.

        An alias for :meth:`~.postal_code()`.

        :return: Zip code.
        """
        return self.postal_code()

    def country_code(self, code: CountryCode | None = CountryCode.A2) -> str:
        """Generates a random code of country.

        Default format is :attr:`~enums.CountryCode.A2` (ISO 3166-1-alpha2),
        you can change it by passing parameter ``fmt``.

        :param code: Country code.
        :return: Country code in selected format.
        :raises KeyError: if fmt is not supported.
        """
        key = self.validate_enum(code, CountryCode)
        return self.random.choice(COUNTRY_CODES[key])

    def country_emoji_flag(self) -> str:
        """Generates a randomly chosen country emoji flag.

        :example:
            🇹🇷

        :return: Flag emoji.
        """
        code = self.country_code(
            code=CountryCode.A2,
        )

        offset = ord("🇦") - ord("A")
        first = ord(code[0]) + offset
        second = ord(code[1]) + offset
        return chr(first) + chr(second)

    def default_country(self) -> str:
        """Returns the country associated with the current locale.

        :return: The country associated with current locale.
        """
        country: str = self._extract(["country", "current_locale"])
        return country

    def country(self) -> str:
        """Generates a random country.

        :return: The Country.
        """
        countries: list[str] = self._extract(["country", "name"])
        return self.random.choice(countries)

    def city(self) -> str:
        """Generates a random city.

        :return: City name.
        """
        cities: list[str] = self._extract(["city"])
        return self.random.choice(cities)

    def _get_fs(self, key: str, dms: bool = False) -> str | float:
        """Get float number.

        :param key: Key (`lt` or `lg`).
        :param dms: DMS format.
        :return: Float number
        """
        # The default range is a range of longitudes.
        rng = (-90, 90) if key == "lt" else (-180, 180)
        result = self.random.uniform(*rng, precision=6)

        if dms:
            return self._dd_to_dms(result, key)

        return result

    def latitude(self, dms: bool = False) -> str | float:
        """Generates a random value of latitude.

        :param dms: DMS format.
        :return: Value of longitude.
        """
        return self._get_fs("lt", dms)

    def longitude(self, dms: bool = False) -> str | float:
        """Generates a random value of longitude.

        :param dms: DMS format.
        :return: Value of longitude.
        """
        return self._get_fs("lg", dms)

    def coordinates(self, dms: bool = False) -> dict[str, str | float]:
        """Generates random geo coordinates.

        :param dms: DMS format.
        :return: Dict with coordinates.
        """
        return {
            "longitude": self._get_fs("lg", dms),
            "latitude": self._get_fs("lt", dms),
        }

    def continent(self, code: bool = False) -> str:
        """Returns a random continent name or continent code.

        :param code: Return code of a continent.
        :return: Continent name.
        """
        codes: list[str] = self._extract(["continent"])

        if code:
            codes = CONTINENT_CODES

        return self.random.choice(codes)

    def calling_code(self) -> str:
        """Generates a random calling code of random country.

        :return: Calling code.
        """
        return self.random.choice(CALLING_CODES)

    def isd_code(self) -> str:
        """Generates a random ISD code.

        An alias for :meth:`~Address.calling_code()`.
        """
        return self.calling_code()



"""Provider of data related to date and time."""

import typing as t
from calendar import monthrange
from datetime import date, datetime, time, timedelta

from mimesis.compat import pytz
from mimesis.datasets import GMT_OFFSETS, ROMAN_NUMS, TIMEZONES
from mimesis.enums import TimestampFormat, TimezoneRegion
from mimesis.providers.base import BaseDataProvider
from mimesis.types import Date, DateTime, Time

__all__ = ["Datetime"]


class Datetime(BaseDataProvider):
    """Class for generating data related to the date and time."""

    # See: https://git.io/Jf15A
    _CURRENT_YEAR = datetime.now().year

    class Meta:
        name = "datetime"
        datafile = f"{name}.json"

    @staticmethod
    def bulk_create_datetimes(
        date_start: DateTime,
        date_end: DateTime,
        **kwargs: t.Any,
    ) -> list[DateTime]:
        """Bulk create datetime objects.

        This method creates a list of datetime objects from
        ``date_start`` to ``date_end``.

        You can use the following keyword arguments:

        * ``days``
        * ``hours``
        * ``minutes``
        * ``seconds``
        * ``microseconds``

        .. warning::

            Empty ``**kwargs`` produces **timedelta(0)** which obviously
            cannot be used as step, so you have to pass valid ``**kwargs``
            for :py:class:`datetime.timedelta` which will be used as a step
            by which ``date_start`` will be incremented until it reaches ``date_end``
            to avoid infinite loop which eventually leads to ``OverflowError``.

        See :py:class:`datetime.timedelta` for more details.

        :param date_start: Begin of the range.
        :param date_end: End of the range.
        :param kwargs: Keyword arguments for :py:class:`datetime.timedelta`
        :return: List of datetime objects
        :raises: ValueError: When ``date_start``/``date_end`` not passed,
            when ``date_start`` larger than ``date_end`` or when the given
            keywords for `datetime.timedelta` represent a non-positive timedelta.
        """
        dt_objects = []

        if not date_start and not date_end:
            raise ValueError("You must pass date_start and date_end")

        if date_end < date_start:
            raise ValueError("date_start can not be larger than date_end")

        if timedelta(**kwargs) <= timedelta():
            raise ValueError("timedelta must be positive")

        while date_start <= date_end:
            date_start += timedelta(**kwargs)
            dt_objects.append(date_start)

        return dt_objects

    def week_date(self, start: int = 2017, end: int = _CURRENT_YEAR) -> str:
        """Generates week number with year.

        :param start: Starting year.
        :param end: Ending year.
        :return: Week number.
        """
        year = self.year(start, end)
        week = self.random.randint(1, 52)
        return f"{year}-W{week}"

    def day_of_week(self, abbr: bool = False) -> str:
        """Generates a random day of the week.

        :param abbr: Abbreviated day name.
        :return: Day of the week.
        """
        key = "abbr" if abbr else "name"
        days: list[str] = self._extract(["day", key])
        return self.random.choice(days)

    def month(self, abbr: bool = False) -> str:
        """Generates a random month of the year.

        :param abbr: Abbreviated month name.
        :return: Month name.
        """
        key = "abbr" if abbr else "name"
        months: list[str] = self._extract(["month", key])
        return self.random.choice(months)

    def year(self, minimum: int = 1990, maximum: int = _CURRENT_YEAR) -> int:
        """Generates a random year.

        :param minimum: Minimum value.
        :param maximum: Maximum value.
        :return: Year.
        """
        return self.random.randint(minimum, maximum)

    def century(self) -> str:
        """Generates a random century.

        :return: Century.
        """
        return self.random.choice(ROMAN_NUMS)

    def periodicity(self) -> str:
        """Generates a random periodicity string.

        :return: Periodicity.
        """
        periodicity: list[str] = self._extract(["periodicity"])
        return self.random.choice(periodicity)

    def date(self, start: int = 2000, end: int = _CURRENT_YEAR) -> Date:
        """Generates a random date object.

        :param start: Minimum value of year.
        :param end: Maximum value of year.
        :return: Formatted date.
        """
        year = self.random.randint(start, end)
        month = self.random.randint(1, 12)
        day = self.random.randint(1, monthrange(year, month)[1])
        date_object = date(year, month, day)
        return date_object

    def formatted_date(self, fmt: str = "", **kwargs: t.Any) -> str:
        """Generates random date as string.

        :param fmt: The format of date, if None then use standard
            accepted in the current locale.
        :param kwargs: Keyword arguments for :meth:`~.date()`
        :return: Formatted date.
        """
        date_obj = self.date(**kwargs)

        if not fmt:
            fmt = self._extract(["formats", "date"])

        return date_obj.strftime(fmt)

    def time(self) -> Time:
        """Generates a random time object.

        :return: ``datetime.time`` object.
        """
        random_time = time(
            self.random.randint(0, 23),
            self.random.randint(0, 59),
            self.random.randint(0, 59),
            self.random.randint(0, 999999),
        )
        return random_time

    def formatted_time(self, fmt: str = "") -> str:
        """Generates formatted time as string.

        :param fmt: The format of time, if None then use standard
            accepted in the current locale.
        :return: String formatted time.
        """
        time_obj = self.time()

        if not fmt:
            fmt = self._extract(["formats", "time"])
        return time_obj.strftime(fmt)

    def day_of_month(self) -> int:
        """Generates a random day of the month, from 1 to 31.

        :return: Random value from 1 to 31.
        """
        return self.random.randint(1, 31)

    def timezone(self, region: TimezoneRegion | None = None) -> str:
        """Generates a random timezone.

        :param region: Timezone region.
        :return: Timezone.
        """
        region_name = self.validate_enum(region, TimezoneRegion)
        return self.random.choice(
            [tz for tz in TIMEZONES if tz.startswith(region_name)]
        )

    def gmt_offset(self) -> str:
        """Generates a random GMT offset value.

        :return: GMT Offset.
        """
        return self.random.choice(GMT_OFFSETS)

    def datetime(
        self,
        start: int = _CURRENT_YEAR,
        end: int = _CURRENT_YEAR,
        timezone: str | None = None,
    ) -> DateTime:
        """Generates random datetime.

        :param start: Minimum value of year.
        :param end: Maximum value of year.
        :param timezone: Set custom timezone (pytz required).
        :return: Datetime
        """
        datetime_obj = datetime.combine(
            date=self.date(start, end),
            time=self.time(),
        )
        if timezone:
            if not pytz:
                raise ImportError("Timezones are supported only with pytz")
            tz = pytz.timezone(timezone)
            datetime_obj = tz.localize(datetime_obj)

        return datetime_obj

    def formatted_datetime(self, fmt: str = "", **kwargs: t.Any) -> str:
        """Generates datetime string in human-readable format.

        :param fmt: Custom format (default is format for current locale)
        :param kwargs: Keyword arguments for :meth:`~.datetime()`
        :return: Formatted datetime string.
        """
        dt_obj = self.datetime(**kwargs)

        if not fmt:
            date_fmt = self._extract(["formats", "date"])
            time_fmt = self._extract(["formats", "time"])
            fmt = f"{date_fmt} {time_fmt}"

        return dt_obj.strftime(fmt)

    def timestamp(
        self, fmt: TimestampFormat = TimestampFormat.POSIX, **kwargs: t.Any
    ) -> str | int:
        """Generates a random timestamp in given format.

        Supported formats are:

        - TimestampFormat.POSIX
        - TimestampFormat.RFC_3339
        - TimestampFormat.ISO_8601

        Example:

        >>> from mimesis import Datetime
        >>> from mimesis.enums import TimestampFormat
        >>> dt = Datetime()
        >>> dt.timestamp(fmt=TimestampFormat.POSIX)
        1697322442
        >>> dt.timestamp(fmt=TimestampFormat.RFC_3339)
        '2023-12-08T18:46:34'
        >>> dt.timestamp(fmt=TimestampFormat.ISO_8601)
        '2009-05-30T21:45:57.328600'

        :param fmt: Format of timestamp (Default is TimestampFormat.POSIX).
        :param kwargs: Kwargs for :meth:`~.datetime()`.
        :return: Timestamp.
        """
        self.validate_enum(fmt, TimestampFormat)
        stamp = self.datetime(**kwargs)

        if fmt == TimestampFormat.RFC_3339:
            return stamp.strftime("%Y-%m-%dT%H:%M:%SZ")
        elif fmt == TimestampFormat.ISO_8601:
            return stamp.isoformat()
        else:
            return int(stamp.timestamp())



"""
This module contains knowledge about the types we use.

Policy
~~~~~~

If any of the following statements is true, move the type to this file:

- if type is used in multiple files
- if type is complex enough it has to be documented
- if type is very important for the public API

"""

import datetime
from decimal import Decimal
from typing import Any, Callable, Final

__all__ = [
    "CallableSchema",
    "Date",
    "DateTime",
    "JSON",
    "Key",
    "Keywords",
    "Matrix",
    "MissingSeed",
    "Seed",
    "Time",
    "Timestamp",
]

JSON = dict[str, Any]

DateTime = datetime.datetime

Time = datetime.time

Date = datetime.date

Timestamp = str | int


class _MissingSeed:
    """We use this type as a placeholder for cases when seed is not set."""


MissingSeed: Final = _MissingSeed()

Seed = None | int | float | str | bytes | bytearray | _MissingSeed

Keywords = list[str] | set[str] | tuple[str, ...]

Number = int | float | complex | Decimal

Matrix = list[list[Number]]

CallableSchema = Callable[[], JSON]

Key = Callable[[Any], Any] | None



"""Implements various helpers which are used in the various data providers.

This module contains custom ``Random()`` class where implemented a lot of
methods which are not included in standard :py:class:`random.Random`,
but frequently used in this project.
"""

import random as random_module
import typing as t

from mimesis.types import MissingSeed, Seed

__all__ = ["Random", "random"]

#: Different plugins (like `pytest-randomly`)
#: can set custom values to a global seed,
#: which are going to be the new default.
global_seed: Seed = MissingSeed


class Random(random_module.Random):
    """A custom random class.

    It is a subclass of the :py:class:`random.Random` class from the standard
    library's random module. The class incorporates additional custom methods.

    This class can be extended according to specific requirements.
    """

    def randints(self, n: int = 3, a: int = 1, b: int = 100) -> list[int]:
        """Generate a list of random integers.

        :param n: Number of elements.
        :param a: Minimum value of range.
        :param b: Maximum value of range.
        :return: List of random integers.
        :raises ValueError: if the number is less or equal to zero.
        """
        if n <= 0:
            raise ValueError("Amount out of range.")

        return [int(self.random() * (b - a)) + a for _ in range(n)]

    def _generate_string(self, str_seq: str, length: int = 10) -> str:
        """Generate random string created from a string sequence.

        :param str_seq: String sequence of letters or digits.
        :param length: Max value.
        :return: Single string.
        """
        return "".join(self.choices(str_seq, k=length))

    def generate_string_by_mask(
        self,
        mask: str = "@###",
        char: str = "@",
        digit: str = "#",
    ) -> str:
        """Generate custom code using ascii uppercase and random integers.

        :param mask: Mask of code.
        :param char: Placeholder for characters.
        :param digit: Placeholder for digits.
        :return: Custom code.
        """
        char_code = ord(char)
        digit_code = ord(digit)

        if char_code == digit_code:
            raise ValueError(
                "The same placeholder cannot be "
                "used for both numbers and characters."
            )

        def random_int(a: int, b: int) -> int:
            b = b - a
            return int(self.random() * b) + a

        _mask = mask.encode()
        code = bytearray(len(_mask))
        for i, p in enumerate(_mask):
            if p == char_code:
                a = random_int(65, 91)  # A-Z
            elif p == digit_code:
                a = random_int(48, 58)  # 0-9
            else:
                a = p
            code[i] = a
        return code.decode()

    def uniform(self, a: float, b: float, precision: int = 15) -> float:
        """Get a random number in the range [a, b) or [a, b] depending on rounding.

        :param a: Minimum value.
        :param b: Maximum value.
        :param precision: Round a number to a given
            precision in decimal digits, default is 15.
        """
        return round(a + (b - a) * self.random(), precision)

    def randbytes(self, n: int = 16) -> bytes:
        """Generate n random bytes."""
        return self.getrandbits(n * 8).to_bytes(n, "little")

    def weighted_choice(self, choices: dict[t.Any, float]) -> t.Any:
        """Returns a random element according to the specified weights.

        :param choices: A dictionary where keys are choices and values are weights.
        :raises ValueError: If choices are empty.
        :return: Random key from dictionary.
        """
        if not choices:
            raise ValueError("Choices cannot be empty.")

        population = list(choices.keys())
        weights = list(choices.values())
        return self.choices(population, weights=weights, k=1)[0]

    def choice_enum_item(self, enum: t.Any) -> t.Any:
        """Get random value of enum object.

        :param enum: Enum object.
        :return: Random value of enum.
        """
        return self.choice(list(enum))


# Compat
# See: https://github.com/lk-geimfari/mimesis/issues/469
random = Random()



"""Specific data provider for Poland (pl)."""

from mimesis.enums import Gender
from mimesis.locales import Locale
from mimesis.providers import BaseDataProvider, Datetime
from mimesis.types import DateTime, MissingSeed, Seed

__all__ = ["PolandSpecProvider"]


class PolandSpecProvider(BaseDataProvider):
    """Class that provides special data for Poland (pl)."""

    def __init__(self, seed: Seed = MissingSeed) -> None:
        """Initialize attributes."""
        super().__init__(locale=Locale.PL, seed=seed)

    class Meta:
        name = "poland_provider"
        datafile = None

    def nip(self) -> str:
        """Generate random valid 10-digit NIP.

        :return: Valid 10-digit NIP
        """
        nip_digits = [int(d) for d in str(self.random.randint(101, 998))]
        nip_digits += [self.random.randint(0, 9) for _ in range(6)]
        nip_coefficients = (6, 5, 7, 2, 3, 4, 5, 6, 7)
        sum_v = sum(nc * nd for nc, nd in zip(nip_coefficients, nip_digits))

        checksum_digit = sum_v % 11
        if checksum_digit > 9:
            return self.nip()
        nip_digits.append(checksum_digit)
        return "".join(map(str, nip_digits))

    def pesel(
        self,
        birth_date: DateTime | None = None,
        gender: Gender | None = None,
    ) -> str:
        """Generate random 11-digit PESEL.

        :param birth_date: Initial birthdate (optional)
        :param gender: Gender of the person.
        :return: Valid 11-digit PESEL
        """
        date_object = birth_date
        if not date_object:
            date_object = Datetime().datetime(1940, 2018)

        date = date_object.date()
        year = date.year % 100
        month = date.month
        day = date.day

        if 1800 <= year <= 1899:
            month += 80
        elif 2000 <= year <= 2099:
            month += 20
        elif 2100 <= year <= 2199:
            month += 40
        elif 2200 <= year <= 2299:
            month += 60

        series_number = self.random.randint(0, 999)

        pesel_digits = list(
            map(int, f"{year:02d}{month:02d}{day:02d}{series_number:03d}")
        )

        if gender == Gender.MALE:
            gender_digit = self.random.choice((1, 3, 5, 7, 9))
        elif gender == Gender.FEMALE:
            gender_digit = self.random.choice((0, 2, 4, 6, 8))
        else:
            gender_digit = self.random.choice(range(10))

        pesel_digits.append(gender_digit)
        pesel_coeffs = (9, 7, 3, 1, 9, 7, 3, 1, 9, 7)
        sum_v = sum(nc * nd for nc, nd in zip(pesel_coeffs, pesel_digits))
        checksum_digit = sum_v % 10
        pesel_digits.append(checksum_digit)
        return "".join(map(str, pesel_digits))

    def regon(self) -> str:
        """Generate random valid 9-digit REGON.

        :return: Valid 9-digit REGON
        """
        regon_coeffs = (8, 9, 2, 3, 4, 5, 6, 7)
        regon_digits = [self.random.randint(0, 9) for _ in range(8)]
        sum_v = sum(nc * nd for nc, nd in zip(regon_coeffs, regon_digits))
        checksum_digit = sum_v % 11
        if checksum_digit > 9:
            checksum_digit = 0
        regon_digits.append(checksum_digit)
        return "".join(map(str, regon_digits))



"""Implements classes for generating data by schema."""

import csv
import inspect
import json
import pickle
import re
from typing import Any, Callable, Sequence

from mimesis.exceptions import (
    AliasesTypeError,
    FieldArityError,
    FieldError,
    FieldNameError,
    FieldsetError,
    SchemaError,
)
from mimesis.locales import Locale
from mimesis.providers.base import BaseProvider
from mimesis.providers.generic import Generic
from mimesis.random import Random
from mimesis.types import JSON, CallableSchema, Key, MissingSeed, Seed

__all__ = [
    "BaseField",
    "Field",
    "Fieldset",
    "Schema",
    "FieldHandler",
    "RegisterableFieldHandler",
    "RegisterableFieldHandlers",
]

FieldCache = dict[str, Callable[[Any], Any]]
FieldHandler = Callable[[Random, Any], Any]
RegisterableFieldHandler = tuple[str, FieldHandler]
RegisterableFieldHandlers = Sequence[RegisterableFieldHandler]


class BaseField:
    def __init__(
        self,
        locale: Locale = Locale.DEFAULT,
        seed: Seed = MissingSeed,
    ) -> None:
        """Base class for fields.

        This class is used as a base class for :class:`Field` and :class:`Fieldset`.

        :attr: aliases: A dictionary of aliases for standard fields.
        :param locale: Locale.
        :param seed: Seed for random.
        """
        self._generic = Generic(locale, seed)
        self._cache: FieldCache = {}
        self._handlers: dict[str, FieldHandler] = {}
        self.aliases: dict[str, str] = {}

    def reseed(self, seed: Seed = MissingSeed) -> None:
        """Reseed the random generator.

        :param seed: Seed for random.
        """
        self._generic.reseed(seed)

    def get_random_instance(self) -> Random:
        """Get a random object from Generic.

        :return: Random object.
        """
        return self._generic.random

    def _explicit_lookup(self, name: str) -> Any:
        """An explicit method lookup.

        This method is called when the field
        defined explicitly, like this: ``provider.method``

        :param name: The field name.
        :return: Callable object.
        :raise FieldError: When field is invalid.
        """
        provider_name, method_name = name.split(".", 1)
        try:
            provider = getattr(self._generic, provider_name)
            return getattr(provider, method_name)
        except AttributeError:
            raise FieldError(name)

    def _fuzzy_lookup(self, name: str) -> Any:
        """A fuzzy method lookup.

        This method is called when the field definition
        is fuzzy, like this: ``method``

        :param name: The field name.
        :return: Callable object.
        :raise FieldError: When field is invalid.
        """
        for provider in dir(self._generic):
            provider = getattr(self._generic, provider)
            if isinstance(provider, BaseProvider):
                if name in dir(provider):
                    return getattr(provider, name)

        raise FieldError(name)

    def _lookup_method(self, name: str) -> Any:
        """Lookup method by the field name.

        :param name: The field name.
        :return: Callable object.
        :raise FieldError: When field is invalid.
        """
        # Check if the field is defined in aliases
        name = self.aliases.get(name, name)

        # Support additional delimiters
        name = re.sub(r"[/:\s]", ".", name)

        if name.count(".") > 1:
            raise FieldError(name)

        if name not in self._cache:
            if "." not in name:
                method = self._fuzzy_lookup(name)
            else:
                method = self._explicit_lookup(name)
            self._cache[name] = method

        return self._cache[name]

    def _validate_aliases(self) -> bool:
        """Validate aliases."""
        if not isinstance(self.aliases, dict) or any(
            not isinstance(key, str) or not isinstance(value, str)
            for key, value in self.aliases.items()
        ):
            # Reset to valid state
            self.aliases = {}
            raise AliasesTypeError()
        return True

    def perform(
        self,
        name: str | None = None,
        key: Key = None,
        **kwargs: Any,
    ) -> Any:
        """Performs the value of the field by its name.

        It takes any string that represents the name of any method of
        any supported data provider and the ``**kwargs`` of this method.

        .. note:: Some data providers have methods with the same names,
            and in such cases, you can explicitly define that the method
            belongs to data-provider ``field(name='provider.name')`` otherwise
            it will return the data from the first provider which
            has a method ``name``.

            Allowed delimiters: ``.``, ``:``, ``/`` and space:

            - ``provider.name``
            - ``provider:name``
            - ``provider/name``
            - ``provider name``

        You can apply a *key function* to the result returned by
        the method, by passing a parameter **key** with a callable
        object which returns the final result.

        The key function has the option to accept two parameters: **result**
        and **random**. In case you require access to a random instance within
        the key function, you must modify the function to accept both of them,
        where the first corresponds to the method result and the second
        corresponds to the instance of random.

        :param name: Name of the method.
        :param key: A key function (any callable object)
            which will be applied to result.
        :param kwargs: Kwargs of method.
        :return: The result of method.
        :raises ValueError: if provider is not supported or if field is not defined.
        """
        # Validate aliases before lookup
        self._validate_aliases()

        if name is None:
            raise FieldError()

        random = self.get_random_instance()

        # First, try to find a custom field handler.
        if name in self._handlers:
            result = self._handlers[name](random, **kwargs)  # type: ignore
        else:
            result = self._lookup_method(name)(**kwargs)

        if key and callable(key):
            try:
                # If a key function accepts two parameters
                # then pass random instance to it.
                return key(result, random)  # type: ignore
            except TypeError:
                return key(result)

        return result

    def register_handler(self, field_name: str, field_handler: FieldHandler) -> None:
        """Register a new field handler.

        :param field_name: Name of the field.
        :param field_handler: Callable object.
        """

        if not isinstance(field_name, str):
            raise TypeError("Field name must be a string.")

        if not field_name.isidentifier():
            raise FieldNameError(field_name)

        if not callable(field_handler):
            raise TypeError("Handler must be a callable object.")

        callable_signature = inspect.signature(field_handler)

        if len(callable_signature.parameters) <= 1:
            raise FieldArityError()

        if field_name not in self._handlers:
            self._handlers[field_name] = field_handler

    def handle(
        self, field_name: str | None = None
    ) -> Callable[[FieldHandler], FieldHandler]:
        """Decorator for registering a custom field handler.

        You can use this decorator only for functions,
        not for any other callables.

        .. versionadded:: 12.0.0

        :param field_name: Name of the field.
            If not specified, the name of the function is used.
        :return: Decorator.
        """

        def decorator(field_handler: FieldHandler) -> FieldHandler:
            _field_name = field_name or field_handler.__name__
            self.register_handler(_field_name, field_handler)
            return field_handler

        return decorator

    def register_handlers(self, fields: RegisterableFieldHandlers) -> None:
        """Register the new field handlers.

        :param fields: A sequence of sequences with field name and handler.
        :return: None.
        """
        for name, handler in fields:
            self.register_handler(name, handler)

    def unregister_handler(self, field_name: str) -> None:
        """Unregister a field handler.

        :param field_name: Name of the field.
        """

        self._handlers.pop(field_name, None)

    def unregister_handlers(self, field_names: Sequence[str] = ()) -> None:
        """Unregister a field handlers with given names.

        :param field_names: Names of the fields.
        :return: None.
        """

        for name in field_names:
            self.unregister_handler(name)

    def unregister_all_handlers(self) -> None:
        """Unregister all custom field handlers.

        :return: None.
        """
        self._handlers.clear()

    def __str__(self) -> str:
        return f"{self.__class__.__name__} <{self._generic.locale}>"


class Field(BaseField):
    """Greedy field (evaluates immediately).

    .. warning::

        There is no case when you need to instance **field** in loops.

        If you are doing this:

        >>> for i in range(1000):
        ...     field = Field()

        You're doing it **wrong**! It is a terrible idea that will lead to a memory leak.

        Forewarned is forearmed.

    Here is an example of how to use it:

        >>> _ = Field()
        >>> _('username')
        Dogtag_1836
    """

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.perform(*args, **kwargs)


class Fieldset(BaseField):
    """Greedy fieldset (evaluates immediately).

    Works like a field, but returns a list of values.

    Here is an example:

        >>> fieldset = Fieldset(i=100)
        >>> fieldset('username')
        ['pot_1821', 'vhs_1915', ..., 'reviewed_1849']

    You may also specify the number of iterations by passing the **i** keyword
    argument to the callable instance of fieldset:

        >>> fieldset = Fieldset()
        >>> fieldset('username', i=2)
        ['pot_1821', 'vhs_1915']

    When **i** is not specified, the reasonable default is used — **10**.

    See "Field vs Fieldset" section of documentation for more details.

    :cvar fieldset_default_iterations: Default iterations. Default is **10**.
    :cvar fieldset_iterations_kwarg: Keyword argument for iterations. Default is **i**.
    """

    fieldset_default_iterations: int = 10
    fieldset_iterations_kwarg: str = "i"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize fieldset.

        Accepts additional keyword argument **i** which is used
        to specify the number of iterations.

        The name of the keyword argument can be changed by
        overriding **fieldset_iterations_kwarg** attribute of this class.
        """
        self._iterations = kwargs.pop(
            self.fieldset_iterations_kwarg,
            self.fieldset_default_iterations,
        )
        super().__init__(*args, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> list[Any]:
        """Perform fieldset.

        :param args: Arguments for field.
        :param kwargs: Keyword arguments for field.
        :raises FieldsetError: If parameter **i** is less than 1.
        :return: List of values.
        """
        min_iterations = 1
        iterations = kwargs.pop(
            self.fieldset_iterations_kwarg,
            self._iterations,
        )

        if iterations < min_iterations:
            raise FieldsetError()

        return [self.perform(*args, **kwargs) for _ in range(iterations)]


class Schema:
    """Class which return list of filled schemas."""

    __slots__ = (
        "__counter",
        "__schema",
        "iterations",
    )

    def __init__(self, schema: CallableSchema, iterations: int = 10) -> None:
        """Initialize schema.

        :param iterations: Number of iterations.
            This parameter is keyword-only. The default value is 10.
        :param schema: A schema (must be a callable object).
        """
        if iterations < 1:
            raise ValueError("Number of iterations should be greater than 1.")

        self.iterations = iterations
        if schema and callable(schema):  # type: ignore[truthy-function]
            self.__schema = schema
            self.__counter = 0
        else:
            raise SchemaError()

    def to_csv(self, file_path: str, **kwargs: Any) -> None:
        """Export a schema as a CSV file.

        :param file_path: The file path.
        :param kwargs: The keyword arguments for :py:class:`csv.DictWriter` class.
        """
        data = self.create()
        with open(file_path, "w", encoding="utf-8", newline="") as fp:
            fieldnames = list(data[0])
            dict_writer = csv.DictWriter(fp, fieldnames, **kwargs)
            dict_writer.writeheader()
            dict_writer.writerows(data)

    def to_json(self, file_path: str, **kwargs: Any) -> None:
        """Export a schema as a JSON file.

        :param file_path: File a path.
        :param kwargs: Extra keyword arguments for :py:func:`json.dump` class.
        """
        with open(file_path, "w", encoding="utf-8") as fp:
            json.dump(self.create(), fp, **kwargs)

    def to_pickle(self, file_path: str, **kwargs: Any) -> None:
        """Export a schema as the pickled representation of the object to the file.

        :param file_path: The file path.
        :param kwargs: Extra keyword arguments for :py:func:`pickle.dump` class.
        """
        with open(file_path, "wb") as fp:
            pickle.dump(self.create(), fp, **kwargs)

    def create(self) -> list[JSON]:
        """Creates a list of a fulfilled schemas.

        .. note::
            This method evaluates immediately, so be careful when creating
            large datasets otherwise you're risking running out of memory.

            If you need a lazy version of this method, see :meth:`iterator`.

        :return: List of fulfilled schemas.
        """
        return [self.__schema() for _ in range(self.iterations)]

    def __next__(self) -> JSON:
        """Return the next item from the iterator."""
        if self.__counter < self.iterations:
            self.__counter += 1
            return self.__schema()
        raise StopIteration

    def __iter__(self) -> "Schema":
        """Return the iterator object itself."""
        self.__counter = 0
        return self



"""Specific data provider for Brazil (pt-br)."""

from mimesis.locales import Locale
from mimesis.providers import BaseDataProvider
from mimesis.types import MissingSeed, Seed

__all__ = ["BrazilSpecProvider"]


class BrazilSpecProvider(BaseDataProvider):
    """Class that provides special data for Brazil (pt-br)."""

    def __init__(self, seed: Seed = MissingSeed) -> None:
        """Initialize attributes."""
        super().__init__(locale=Locale.PT_BR, seed=seed)

    class Meta:
        name = "brazil_provider"
        datafile = None

    @staticmethod
    def __get_verifying_digit_cpf(cpf: list[int], weight: int) -> int:
        """Calculate the verifying digit for the CPF.

        :param cpf: List of integers with the CPF.
        :param weight: Integer with the weight for the modulo 11 calculate.
        :returns: The verifying digit for the CPF.
        """
        total = 0

        for index, digit in enumerate(cpf):
            total += digit * (weight - index)

        remainder = total % 11

        if remainder == 0 or remainder == 1 or remainder >= 11:
            return 0

        return 11 - remainder

    def cpf(self, with_mask: bool = True) -> str:
        """Get a random CPF.

        :param with_mask: Use CPF mask (###.###.###-##).
        :returns: Random CPF.

        :Example:
            001.137.297-40
        """

        cpf_without_dv = [self.random.randint(0, 9) for _ in range(9)]
        first_dv = self.__get_verifying_digit_cpf(cpf_without_dv, 10)

        cpf_without_dv.append(first_dv)
        second_dv = self.__get_verifying_digit_cpf(cpf_without_dv, 11)
        cpf_without_dv.append(second_dv)

        cpf = "".join(str(i) for i in cpf_without_dv)

        if with_mask:
            return f"{cpf[:3]}.{cpf[3:6]}.{cpf[6:9]}-{cpf[9:]}"
        return cpf

    @staticmethod
    def __get_verifying_digit_cnpj(cnpj: list[int], weight: int) -> int:
        """Calculate the verifying digit for the CNPJ.

        :param cnpj: List of integers with the CNPJ.
        :param weight: Integer with the weight for the modulo 11 calculate.
        :returns: The verifying digit for the CNPJ.
        """
        total = 0
        weights_dict = {
            5: [5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2],
            6: [6, 5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2],
        }
        weights = weights_dict[weight]

        for i, _ in enumerate(cnpj):
            total += weights[i] * cnpj[i]

        remainder = total % 11
        return 0 if (remainder < 2) else (11 - remainder)

    def cnpj(self, with_mask: bool = True) -> str:
        """Get a random CNPJ.

        :param with_mask: Use cnpj mask (###.###.###-##)
        :returns: Random cnpj.

        :Example:
            77.732.230/0001-70
        """

        cnpj_without_dv = [self.random.randint(0, 9) for _ in range(12)]

        first_dv = self.__get_verifying_digit_cnpj(cnpj_without_dv, 5)
        cnpj_without_dv.append(first_dv)

        second_dv = self.__get_verifying_digit_cnpj(cnpj_without_dv, 6)
        cnpj_without_dv.append(second_dv)

        cnpj = "".join(str(i) for i in cnpj_without_dv)

        if with_mask:
            return "{}.{}.{}/{}-{}".format(
                cnpj[:2], cnpj[2:5], cnpj[5:8], cnpj[8:12], cnpj[12:]
            )
        return cnpj



"""Data related to the development."""

import typing as t
from datetime import datetime

from mimesis.datasets import (
    LICENSES,
    OS,
    PROGRAMMING_LANGS,
    STAGES,
    SYSTEM_QUALITY_ATTRIBUTES,
)
from mimesis.providers.base import BaseProvider

__all__ = ["Development"]


class Development(BaseProvider):
    """Class for getting fake data for Developers."""

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

    class Meta:
        name = "development"

    def software_license(self) -> str:
        """Generates a random software license.

        :return: License name.

        :Example:
            The BSD 3-Clause License.
        """
        return self.random.choice(LICENSES)

    def calver(self) -> str:
        """Generates a random calendar versioning string.

        :return: Calendar versioning string.

        :Example:
            2016.11.08
        """
        year = self.random.randint(2016, datetime.now().year)
        month = self.random.randint(1, 12)
        day = self.random.randint(1, 29)
        return f"{year}.{month}.{day}"

    def version(self) -> str:
        """Generates a random semantic versioning string.

        :return: Semantic versioning string.

        :Example:
            0.2.1
        """
        major, minor, patch = self.random.randints(n=3, a=0, b=100)
        return f"{major}.{minor}.{patch}"

    def stage(self) -> str:
        """Generates a random stage of development.

        :return: Release stage.

        :Example:
            Alpha.
        """
        return self.random.choice(STAGES)

    def programming_language(self) -> str:
        """Generates a random programming language from the list.

        :return: Programming language.

        :Example:
            Erlang.
        """
        return self.random.choice(PROGRAMMING_LANGS)

    def os(self) -> str:
        """Generates a random operating system or distributive name.

        :return: The name of OS.

        :Example:
            Gentoo
        """
        return self.random.choice(OS)

    def boolean(self) -> bool:
        """Generates a random boolean value.

        :return: True of False.
        """
        return self.random.choice([True, False])

    def system_quality_attribute(self) -> str:
        """Generates a random system quality attribute.

        Within systems engineering, quality attributes are realized
        non-functional requirements used to evaluate the performance
        of a system. These are sometimes named "ilities" after the
        suffix many of the words share.

        :return: System quality attribute.
        """
        return self.random.choice(SYSTEM_QUALITY_ATTRIBUTES)

    def ility(self) -> str:
        """Generates a random system quality attribute.

        An alias for :meth:`~mimesis.Development.system_quality_attribute`.
        """
        return self.system_quality_attribute()
